import collections
from slugify import slugify
import itertools
import logging
import operator
import typing
from pathlib import Path

import more_itertools

from .feeds import RSSFeed
from .page import Page


class Collection:
    """Collection objects serve as a way to quickly process pages that have a
    LARGE portion of content that is similar or file driven.

    The most common form of collection would be a Blog, but can also be
    static pages that have their content stored in a dedicated file.

    Currently, collections must come from a content_path and all be the same
    content type.


    Example
    -------
    from render_engine import Collection

    @site.register_collection()
    class BasicCollection(Collection):
        pass


    Attributes
    ----------
    engine: str, optional
        The engine that the collection will pass to each page. Site's default
        engine
    template: str
        The template that each page will use to render
    routes: List[str]
        all routes that the file should be created at. default []
    content_path: List[PathString], optional
        the filepath to load content from.
    includes: List[str], optional
        the types of files in the content path that will be processed
        default ["*.md", "*.html"]
    has_archive: Bool
        if `True`, create an archive page with all of the processed pages saved
        as `pages`. default `False`
    template: str, optional
        template filename that will be used if `has_archive==True` default: archive.html"
    slug: str, optional
        slug for rendered page if `has_archive == True` default: all_posts
    content_type: Type[Page], optional
        content_type for the rendered archive page
    _archive_reverse: Bool, optional
        should the sorted `pages` be listed in reverse order. default: False

    """

    engine: typing.Optional[str] = None
    content_items: typing.List[Page] = []
    content_path: str = ""
    content_type: typing.Type[Page] = Page
    template: str = "page.html"
    includes: typing.List[str] = ["*.md", "*.html"]
    routes: typing.List[str] = [""]
    subcollections: typing.List[str] = []
    has_archive: bool = False
    archive_template: str = "archive.html"
    archive_reverse: bool = False
    archive_sort: typing.Tuple[str] = "title"
    paginated: bool = False
    items_per_page: int = 10
    title: typing.Optional[str] = ""
    feeds: typing.List[typing.Optional[RSSFeed]] = []
    markdown_extras = ["fenced-code-blocks", "footnotes"]

    def __init__(self):
        if not self.title:
            self.title = self.__class__.__name__

    @property
    def slug(self):
        return slugify(self.title)

    @property
    def pages(self) -> typing.List[Page]:
        _pages = []

        if self.content_items:
            _pages = self.content_items

        if self.content_path:

            if Path(self.content_path).samefile("/"):
                logging.warning(
                    f"{self.content_path=}! Accessing Root Directory is Dangerous..."
                )

            for pattern in self.includes:

                for filepath in Path(self.content_path).glob(pattern):
                    page = self.content_type.from_content_path(
                        filepath,
                        markdown_extras=self.markdown_extras,
                    )
                    page.routes = self.routes
                    page.template = self.template
                    _pages.append(page)

        return _pages

    @property
    def archive(self):
        """Create a `Page` object for the pages in the collection"""

        sorted_pages = sorted(
            self.pages,
            key=lambda p: getattr(p, self.archive_sort),
            reverse=self.archive_reverse,
        )

        if self.paginated:
            pages = list(more_itertools.chunked(sorted_pages, self.items_per_page))

        else:
            pages = [sorted_pages]

        class Archive(Page):
            no_index = True
            template = self.archive_template
            routes = [self.routes[0]]
            title = self.title

        archive_pages = []

        for index, page in enumerate(pages):
            archive_page = Archive()
            archive_page.collection = self
            archive_page.routes = [self.routes[0]]
            archive_page.pages = pages[index]
            archive_page.title = self.title
            archive_page.page_index = (index, len(pages))

            if self.paginated:
                archive_page.slug = f"{archive_page.slug}-{index}"

            archive_pages.append(archive_page)

        return archive_pages

    def get_subcollections(self):
        subcollections = {}

        # get all the values for each of the subcollections
        for _subcollection in self.subcollections:
            subcollection_lists = collections.defaultdict(list)
            subcollections[_subcollection] = []

            for page in self.pages:

                if attr := getattr(page, _subcollection, None):

                    if isinstance(attr, list):

                        for attribute_item in attr:
                            subcollection_lists[attribute_item].append(page)

                    else:
                        subcollection_lists[attr].append(page)

            for attr, subcollection in subcollection_lists.items():

                class SubCollection(self.__class__):
                    content_path = None
                    content_items = subcollection
                    title = attr
                    slug = slugify(attr)

                subcollections[_subcollection].append(SubCollection())

        return subcollections



from .content_parser import parse_content
from slugify import slugify
import logging
import more_itertools
import typing
from pathlib import Path
from typing import List

from jinja2 import Markup
from markdown2 import markdown

from ._type_hint_helpers import PathString

# some attributes will need to be protected from manipulation


class Page:
    """
    Base component used to make web pages.

    All components that represent content in HTML, XML, or JSON generated by
    Render Engine should be a Page object.

    Pages can be rendered directly from a template or from a file.

    Page objects can be used to extend existing page objects.


    Examples:
        ```
        # Basic Page with No Template Variables
        @site.register_route('basic_page.html')
        class BasicPage(Page):
            template = 'template_file.html' # user provided template

        # Basic Page with Variables
        @site.register_route('page_with_vars')
        class PageWithVars(Page):
            title = 'Site Title'

        # Page Loading from File
        @site.register_route('page_from_file')
        class PageFromFile(Page):
            content_path = 'index.md' # loaded from content path can be '.md' or '.html'

        # Page Inherited from Other Page
        @site.register_route('basic_page.html')
        class BasicPage(Page):
            template = 'template_file.html' # user provided template
            title = 'Base Page'

        @site.register_route('other_page.html')
        class InheritingPage(Page):
            # template will be inherited from the BasicPage
            title = 'Inherited Page'
        ```

    Attributes:
        engine: str, optional
            The engine that the Site should refer to or the site's default engine
        template: str
            The template that the Site should refer to. If empty, use site's default
        routes: List[str]
            all routes that the file should be created at. default []
        content_path: List[PathString], optional
            the filepath to load content from.
        slug: str
            The engine's default route filename
        content: str
            preprocessed text stripped from initialized content. This will not
            include any defined attrs
        html: str
            text converted to html from _content
    """

    engine: typing.Optional[str] = ""  # (default) inherits from Site
    template: typing.Optional[str] = ""  # (default) inherits from Site
    routes = [""]  # create page at each route
    list_attrs: typing.Optional[typing.Union[str]] = []  # comma-delim listed attrs
    no_index: bool = False  # hides from search index
    matcher: str = r"(^\w+: \b.+$)"  # expression to find attrs/vals
    content_path: typing.Optional[str] = ""
    base_content: typing.Optional[str] = ""
    title: typing.Optional[str] = ""
    slug: typing.Optional[str] = ""
    markdown_extras = ["fenced-code-blocks", "footnotes"]

    def __init__(
        self,
    ):

        if self.content_path:
            content = Path(self.content_path).read_text()

            valid_attrs, self.base_content = parse_content(
                content,
                matcher=self.matcher,
            )

            for attr in valid_attrs:
                name, value = attr.split(": ", maxsplit=1)

                # comma delimit attributes.
                if name.lower() in self.list_attrs:
                    value = [attrval.lower() for attrval in value.split(", ")]

                else:
                    value = value.strip()

                setattr(self, name.lower(), value)

        if not self.title:
            self.title = self.__class__.__name__

        if not self.slug:
            self.slug = self.title or self.__class__.__name__

        self.slug = slugify(self.slug)

    @property
    def url(self):
        return f"{self.routes[0]}/{self.slug}"

    @classmethod
    def from_content_path(cls, filepath, **kwargs):
        class NewPage(cls):
            content_path = filepath

            def __init__(self, markdown_extras: typing.List[str] = []):
                for extra in markdown_extras:
                    if extra not in self.markdown_extras:
                        self.markdown_extras.append(extra)

                super().__init__()

        newpage = NewPage(**kwargs)

        return newpage

    @property
    def html(self):
        """Text from self.content converted to html"""

        if self.base_content:
            return markdown(self.base_content, extras=self.markdown_extras)

        else:
            return ""

    @property
    def markup(self):
        """html = rendered html (not marked up). Is `None` if `content == None`"""
        if self.html:
            return Markup(self.html)

        else:
            return ""

    @property
    def content(self):
        return self.markup



import logging
import typing

import pendulum

from .blog import Blog, BlogPost
from .feeds import RSSFeedEngine, RSSFeedItem


class MicroBlogPost(BlogPost):
    """
    Page Like Object with slight modifications to work with BlogPosts

    Attribtues:
        title : str
            default ''. Leave blank.

        slug : str
            the name for the file for that will

        rss_feed_item : RSSFeedItem
            the content in an rss format
    """

    def __init__(self, **kwargs):
        """checks published options and accepts the first that is listed"""
        super().__init__(**kwargs)

        self.slug = self.date_published.format("YMMDDHHmmss")
        self.title = ""


class MicroBlog(Blog):
    """
    Custom Blog Class pointing to custom templates
    """

    _archive_template = "microblog_archive.html"
    _archive_slug = "all_microblog_posts"
    content_type: typing.Type[MicroBlogPost] = MicroBlogPost
    _archive_reverse: bool = True
    has_archive = True



from copy import copy
from progress.bar import Bar
import itertools
import inspect
import logging
import os
import shutil
import typing
import pendulum
from pathlib import Path
from slugify import slugify


import more_itertools

from ._type_hint_helpers import PathString
from .collection import Collection
from .engine import Engine
from .feeds import RSSFeedEngine
from .links import Link
from .page import Page
from . import search


class Site:
    """
    The site stores your pages and collections to be rendered.

    Pages are stored and created with `site.render()`.
    Collections are stored to be used for future use.
    Sites also contain global variables that can be applied in templates.

    Attributes:
        routes (list):
            storage of registered_routes
        collections (dict):
            storage of registered collections
        output_path (str or pathlib.Path):
            the path to directory which all rendered html pages will be stored.
            default `./output`
        static_path (str or pathlib.Path):
            the path to directory for static content. This will be copied over
            into the `output_path`
        SITE_TITLE (str):
            configuration variable title of the site. This is only used in your
            environment template variables. While Optional you will be warned
            if you do not supply a new variable. default: 'Untitled Site'
        SITE_URL (str):
            configuration variable url of the of the site. While Optional you will be
            warned if you do not supply a new variable. default: 'Untitled Site'
            default 'https://example.com'

    Todo:
        - remove SITE_LINK
        - make SITE_URL accesible as a Page variable and allow for switch for
            Relative and Absolute URLS
    """

    routes: typing.List[Page] = []
    output_path: Path = Path("output")
    static_path: Path = Path("static")
    SITE_TITLE: str = "Untitled Site"
    SITE_URL: str = "https://example.com"
    strict: bool = False
    default_engine: typing.Type[Engine] = Engine()
    rss_engine: typing.Type[Engine] = RSSFeedEngine()
    search: typing.Optional[str] = None
    search_params: typing.List[str] = {}
    search_client = None
    timezone: str = ""

    def __init__(self):
        """Clean Directory and Prepare Output Directory"""

        self.collections = {}
        self.subcollections = {}
        self.output_path = Path(self.output_path)

        # strict deletes the directory and rebuilds it from scratch
        if self.output_path.exists() and not self.output_path.is_dir():
            raise ValueError("output path cannot point to an existing file")

        # sets the timezone environment variable to the local timezone if not present
        os.environ["render_engine_timezone"] = (
            self.timezone or pendulum.local_timezone().name
        )

    def register_collection(self, collection_cls: typing.Type[Collection]) -> None:
        """
        Add a class to your `self.collections`
        iterate through a classes `content_path` and create a classes
        `Page`-like objects, adding each one to `routes`.

        Use a decorator for your defined classes.

        Examples:
            ```
            @register_collection
            class Foo(Collection):
                pass
            ```

        Args:
            collection_cls (Collection):
                Collection to parse
        """
        collection = collection_cls()
        self.collections.update({collection.title: collection})

        for page in collection.pages:
            self.routes.append(page)

        if collection.has_archive:

            for archive in collection.archive:
                self.routes.append(archive)

        if collection.feeds:

            for feed in collection.feeds:
                self.register_feed(feed=feed, collection=collection)

    def register_feed(self, feed, collection: Collection) -> None:
        """Create a Page object that is an RSS feed and add it to self.routes

        Parameters
        ----------
        feed: RSSFeedEngine
            the type of feed to generate
        collection: Collection
            the collection to

        Returns
        -------
        None
        """

        extension = self.rss_engine.extension
        _feed = feed
        _feed.slug = collection.slug
        _feed.title = f"{self.SITE_TITLE} - {_feed.title}"
        _feed.link = f"{self.SITE_URL}/{_feed.slug}{extension}"
        self.routes.append(_feed)

    def register_route(self, cls) -> None:
        route = cls()
        self.routes.append(route)

    def _remove_output_path(self):
        if self.output_path.exists():
            return shutil.rmtree(self.output_path)

    def _render_output(self, page):
        """Writes the page to a file"""
        engine = page.engine if page.engine else self.default_engine
        template_attrs = self.get_public_attributes(page)
        content = engine.render(page, **template_attrs)
        route = self.output_path.joinpath(page.routes[0].strip("/"))
        route.mkdir(exist_ok=True)
        filename = Path(page.slug).with_suffix(engine.extension)
        filepath = route.joinpath(filename)
        filepath.write_text(content)
        route_count = len(page.routes)

        if route_count > 1:

            # create a directory and path for each alternate route
            for new_route in page.routes[1:]:
                new_route = self.output_path.joinpath(new_route.strip("/"))
                new_route.mkdir(exist_ok=True)
                new_filepath = new_route.joinpath(filename)
                shutil.copy(filepath, new_filepath)

    def _render_subcollections(self):
        """Generate subcollection pages to be added to routes"""
        for _, collection in self.collections.items():

            if collection.subcollections:

                for subcollection_group in collection.get_subcollections():
                    _subcollection_group = collection.get_subcollections()[
                        subcollection_group
                    ]
                    sorted_group = sorted(
                        _subcollection_group,
                        key=lambda x: (len(x.pages), x.title),
                        reverse=True,
                    )

                    for subcollection in sorted_group:

                        self.subcollections[subcollection_group] = sorted_group

                        for archive in subcollection.archive:
                            self.routes.append(archive)

    def render(self, dry_run: bool = False, strict: bool = False) -> None:
        # removes the output path is strict is set
        if self.strict or strict:
            self._remove_output_path

        # create an output_path if it doesn't exist
        self.output_path.mkdir(exist_ok=True)

        # copy a defined static path into output path
        if Path(self.static_path).is_dir():
            shutil.copytree(
                self.static_path,
                self.output_path.joinpath(self.static_path),
                dirs_exist_ok=True,
            )

        # render registered subcollections
        self._render_subcollections()
        page_count = len(self.routes)

        with Bar(
            f"Rendering {page_count} Pages",
            max=page_count,
            suffix="%(percent).1f%% - %(elapsed_td)s",
        ) as bar:

            for page in self.routes:
                suffix = "%(percent).1f%% - %(elapsed_td)s"
                bar.suffix = suffix + f" ({page.title})"

                self._render_output(page)

                bar.next()

        if self.search:
            search_fields = {
                "title": {
                    "type": "text",
                },
                "content": {
                    "type": "text",
                },
                "slug": {
                    "type": "text",
                },
                "date_published": {
                    "type": "date",
                },
                "date_modified": {
                    "type": "date",
                },
                "tags": {
                    "type": "text",
                    "default": [""],
                },
                "category": {
                    "type": "keyword",
                },
            }
            self.search_params["id_fields"] = ["slug", "date_created"]
            self.search_params["fields"] = search_fields
            filtered_routes = itertools.filterfalse(lambda x: x.no_index, self.routes)
            self.search(
                search_client=self.search_client,
                pages=filtered_routes,
                **self.search_params,
            )

    def get_public_attributes(self, cls):
        site_filtered_attrs = itertools.filterfalse(
            lambda x: x[0].startswith("__"), inspect.getmembers(self)
        )
        site_dict = {x: y for x, y in site_filtered_attrs}

        cls_filtered_attrs = itertools.filterfalse(
            lambda x: x[0].startswith("__"), inspect.getmembers(cls)
        )

        cls_dict = {x: y for x, y in cls_filtered_attrs}

        return {**site_dict, **cls_dict}



import re


def parse_content(content: str, matcher: str):
    """
    split content into attributes and content text

    Parameters:
        content : str
            The content to be parsed
        matcher : str, optional
            A compiled regular expression that splits the content.
            default `base_matcher`
    """

    matchmaker = re.compile(matcher, flags=re.M)
    split_content = content.split("\n\n", maxsplit=1)
    attr_section = split_content[0]

    if len(split_content) == 2:
        base_content = split_content[1]

    else:
        base_content = ""

    parsed_attrs = re.split(matchmaker, attr_section)
    content = base_content.strip()

    attrs = list(filter(lambda x: x.strip(), parsed_attrs))
    return attrs, content



import os
import logging
import typing
from typing import List

import pendulum
import more_itertools

from .collection import Collection
from .feeds import RSSFeed, RSSFeedEngine, RSSFeedItem
from .page import Page


class BlogPost(Page):
    """
    Page Like object with slight modifications to work with BlogPosts.

    Attributes:
        template : str
            the default template that the site will look for
        rss_feed_item : RSSFeedItem
            the content in an rss format
        date : pendulum.datetime
            date parsed in datetime format. usesul for sorting and things
        date_published : str
            date formated for `RSSFeed`
        date_friendly : str
            an easy to read string version of the date

    """

    template = "blog_post.html"
    list_attrs = ["tags"]

    def __init__(self, **kwargs):
        """
        checks published options and accepts the first that is listed

        Attributes:
            date : pendulum.datetime
                date parsed in datetime format. usesul for sorting and things
            date_published : str
                date formated for `RSSFeed`
            date_friendly : str
                an easy to read string version of the date
        """

        super().__init__(**kwargs)

        date_published = more_itertools.first_true(
            (
                getattr(self, "date_published", None),
                getattr(self, "publish_date", None),
                getattr(self, "date", None),
            )
        )

        self.date_published = pendulum.parse(date_published, strict=False).set(
            tz=pendulum.local_timezone()
        )

        date_modified = more_itertools.first_true(
            (
                getattr(self, "date_modified", None),
                getattr(self, "modified_date", None),
            ),
            default=None,
        )

        if not date_modified:
            self.date_modified = self.date_published

        else:
            self.date_modified = pendulum.parse(date_modified, strict=False).set(
                tz=pendulum.local_timezone()
            )

        self.date_friendly = self.date_modified.format("MMM DD, YYYY HH:mm A")

    @property
    def rss_feed_item(self):
        return RSSFeedItem(self)


class Blog(Collection):
    """
    Custom Collection Class with Archiving Enabled and the RSS Feed

    Todos:
        - Add Support for JSON Feeds
        - Rename the archive items so they are not private
    """

    content_type: typing.Type[BlogPost] = BlogPost
    archive_reverse: bool = True
    has_archive: bool = True
    archive_sort = "date"

    @property
    def feeds(self):
        return [RSSFeed(collection=self, title=self.title)]



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
# from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='render_engine',  # Required
    version='2020.10.2',  # Required
    description='Static Page Generation with Flask-like simplicity and flair ✨',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://render-engine.site',  # Optional
    author='Jay Miller',  # Optional
    author_email='jay@productivityintech.com',  # Optional
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Text Processing :: Markup :: HTML',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Natural Language
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='static site web framework',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
    python_requires='>=3.8',
    install_requires=[
    'pendulum',
    'pygments',
    'markdown2',
    'progress',
    'jinja2',
    'more-itertools',
    'click',
    'python-slugify',
    'elasticsearch',
    'elastic_app_search',
    ],  # Optional
    extras_require={  # Optional
        'dev': ['check-manifest', 'elasticsearch', 'elastic_app_search'],
        'test': ['pytest', 'pytest-cov', 'black'],
    },

    package_data={  # Optional
    #    'sample': ['package_data.dat'],
         'render_engine': ['rss/*', 'templates/*', 'run_template.txt'],
    },
    entry_points={  # Optional
        'console_scripts': [
            'render-engine=render_engine.__main__:main',
            'render-engine-quickstart=render_engine.quickstart:quickstart',
        ],
    },
    project_urls={  # Optional
        'Documentation': 'https://render-engine.readthedocs.io/en/latest',
        'Bug Reports': 'https://github.com/kjaymiller/render_engine/issues',
        'Funding': 'https://paypal.me/kjaymiller',
        'Say Thanks!': 'http://saythanks.io/to/kjaymiller',
        'Source': 'https://github.com/kjaymiller/render_engine/',
    },
)



from .page import Page
from elasticsearch.helpers import bulk
from dataclasses import dataclass
import elastic_app_search as app_search


import elasticsearch

from functools import wraps
import json
import logging
import typing
import hashlib


def _id_hash(page, *id_fields):
    """Generate a Hash from the provided fields"""
    raw_msg = "".join([getattr(page, x, "") for x in id_fields]).encode("utf-8")
    return hashlib.sha1(raw_msg).hexdigest()


def _build_index(pages, id_field="_id", **search_params):
    """Build a page dict for the keys requested. There is a
    question if this should be replaced with a __dict__ overwrite on page."""

    for page in pages:
        page_dict = {}

        for key, key_params in search_params["fields"].items():

            if field_value := getattr(page, key, None):

                if key_params["type"] == "date":
                    page_dict[key] = field_value.to_rfc3339_string()

                else:
                    page_dict[key] = field_value

        page_dict[id_field] = _id_hash(page, *search_params["id_fields"])

        yield page_dict


def fuse(search: typing.Dict, filepath: str):
    """Build a JSON doc of your pages"""

    with open(filepath, "w") as jsonfile:
        return json.dump(
            [x for x in _build_index(search, id_field="id")],
            fp=jsonfile,
        )


def elasticsearch(
    _,  # called inside a class
    search_client: elasticsearch.Elasticsearch,
    pages,
    **search_params,
):
    """Upload the pages to an elasticsearch index

    Parameters
    ----------
    search: Search-like object
        Search object from site to push to elasticsearch instance
    pages: list[Page]
        page-like objects to build_indexes out of

    Returns
    -------
    bulk operation
        the bulk command sent to your elasticsearch instance"""
    return bulk(
        client=search_client,
        index=search_params["index"],
        actions=(x for x in _build_index(pages=pages, **search_params)),
    )


def elastic_app_search(
    _,
    search_client,
    pages,
    **search_params,
):
    """Upload the pages to an enterprise app search engine

    Parameters
    ----------
    host: str
        host for the enterprise app search instance.
    api_key: str
        **private** api_key for the app search instance
    engine: str
        name of the app search engine to add documents to
    pages: list[Page]
        page-like objects to build_indexes out of
    use_https: bool, default=False
        is the host using https
    fields: fields to iniclude in the build_index
    id_fields: the fields to use in generating the id (to prevent duplicate entries)
    """

    documents = [x for x in _build_index(pages=pages, id_field="id", **search_params)]
    return search_client.index_documents(
        engine_name=search_params["engine"],
        documents=documents,
    )
